==> lhcsmapi-1.3.300.tar.gz: lhcsmapi-1.3.300/lhcsmapi/reference/Reference.py <==
import os
from typing import List

import pandas as pd
from dateutil import parser
from IPython import get_ipython
from IPython.display import display, HTML

from lhcsmapi.Time import Time


selected_path = os.path.dirname(__file__)


class Reference(object):
    """Reference class provides methods to access tables with reference values of components:

    - EE
    - Current Leads
    - MP3
    - Power Converter
    - QH
    - Query Table

    """

    # Query MP3 Quench Excel database columns
    @staticmethod
    def read_quench_database_columns(circuit_type: str, table_type='MP3') -> List[str]:
        """ Method reading columns of an MP3 quench database Excel file

        :param circuit_type: circuit type
        :type circuit_type: str
        :param table_type: table type (either MP3 or LHCSM)
        :type table_type: str

        :return: pd.DataFrame with names of columns for an MP3 quench database Excel file
        """

        full_path = os.path.join(selected_path, 'mp3/%s_%s_Quench_Database_Columns.csv' % (circuit_type, table_type))
        return pd.read_csv(full_path, index_col=0).columns

    # QH
    @staticmethod
    def read_quench_heater_reference_discharges(circuit_type: str) -> pd.DataFrame:
        """ Method reading reference quench heater discharge timestamps for a given circuit type

        :param circuit_type: circuit type
        :type circuit_type: str

        :return: pd.DataFrame with names of columns for full FPA quench analysis
        """

        full_path = os.path.join(selected_path, 'qh/{}_Reference_Discharges.csv'.format(circuit_type))
        return pd.read_csv(full_path, index_col=0)

    @staticmethod
    def get_quench_heater_reference_discharge(circuit_type: str, cell_or_circuit: str) -> int:
        """Method returning a timestamp corresponding to a reference quench heater discharge

        :param circuit_type:
        :type circuit_type: str
        :param cell_or_circuit: name of a magnet cell
        :type cell_or_circuit: str

        :return: int in a nanosecond precision unix timestamp of a reference QH timestamp
        """
        qh_ref_discharges = Reference.read_quench_heater_reference_discharges(circuit_type)
        if circuit_type in ['RB', 'RQ']:
            col_name = 'Cell'
        elif circuit_type in ['IT', 'IPD2', 'IPD2_B1B2', 'IPQ']:
            col_name = 'Circuit'
        else:
            raise ValueError('Circuit type %s not supported!' % circuit_type)
        
        return qh_ref_discharges[qh_ref_discharges[col_name] == cell_or_circuit]["Timestamp"].values[0]

    @staticmethod
    def read_quench_heaters_reference_features(circuit_type: str) -> pd.DataFrame:
        """ Method reading a pandas DataFrame with reference values of QH features
        For all circuits:

        - initial voltage
        - final voltage

        :param circuit_type: circuit type
        :type circuit_type: str

        :return: pd.DataFrame with reference QH features
        """
        full_path = os.path.join(selected_path, 'qh/{}_Reference_Features.csv'.format(circuit_type))
        return pd.read_csv(full_path, index_col=0)

    @staticmethod
    def read_quench_heaters_reference_differences(circuit_type: str) -> pd.DataFrame:
        """ Method reading a pandas DataFrame with allowed differences of QH features from features calculated with
        the reference discharges.
        For all circuits (with QH voltage measurement):

        - voltage time constant

        For circuits with QH current measurement:

        - current time constant
        - initial resistance value

        :param circuit_type: circuit type
        :type circuit_type: str

        :return: pd.DataFrame with allowed differences of QH features from features calculated with the reference
        discharges.
        """
        full_path = os.path.join(selected_path, 'qh/{}_Reference_Differences.csv'.format(circuit_type))
        return pd.read_csv(full_path, index_col=0)

    # EE
    @staticmethod
    def read_energy_extraction_reference_features(circuit_type: str, timestamp_fgc='2021-01-01 00:00:00.000+01:00') \
            -> pd.DataFrame:
        """ Method reading a pandas DataFrame with min/max reference EE features
        RB:

        - R_1
        - R_2
        - t_delay_1
        - t_delay_2
        - tau_u_dump_res_1
        - tau_u_dump_res_2

        RQ:

        - R
        - t_delay
        - tau_u_dump_res

        :param circuit_type: circuit type
        :type circuit_type: str
        :param timestamp_fgc: FGC timestamp needed to select an appropriate EE reference features
        :type timestamp_fgc: str

        :return: pd.DataFrame with reference EE features
        """

        if circuit_type == 'RQ':
            full_path = os.path.join(selected_path, 'ee/RQ_Reference_Features.csv')
        elif circuit_type == 'RB':
            date_change = parser.parse('2018-01-01 00:00:00.000+01:00')
            timestamp_fgc_datetime = Time.to_datetime(timestamp_fgc)

            if timestamp_fgc_datetime > date_change:
                full_path = os.path.join(selected_path, 'ee/RB_Reference_Features_2018_01_01.csv')
            else:
                full_path = os.path.join(selected_path, 'ee/RB_Reference_Features_2014_01_01_2018_01_01.csv')
        else:
            raise KeyError('Circuit type {} not supported!'.format(circuit_type))

        return pd.read_csv(full_path, index_col=0)

    # PC
    @staticmethod
    def read_power_converter_reference_features(circuit_type: str) -> pd.DataFrame:
        """Method reading a pandas DataFrame with min/max reference features of a Power Converter:
        - tau_i_meas

        :param circuit_type: circuit type
        :type circuit_type: str

        :return: pd.DataFrame with reference Power Converter features
        """
        if circuit_type not in ['RB', 'RQ']:
            raise ValueError('Reference features for {} circuit type are not defined.'.format(circuit_type))

        full_path = os.path.join(selected_path, 'pc/{}_Reference_Features.csv'.format(circuit_type))
        return pd.read_csv(full_path, index_col=0)

    @staticmethod
    def get_power_converter_reference_fpa(circuit_type: str, circuit_name: str, col: str) -> str:
        """ Methor returning a timestamp of a reference FGC/EE timestamp for PM database event

        :param circuit_type: circuit type
        :type circuit_type: str
        :param circuit_name: circuit name
        :type circuit_name: str
        :param col: column: RB: fgcPm,eeEvenPm,eeOddPm, RQ: fgcPm,eePm
        :type col: str

        :return: str with a timestamp of a reference FGC/EE timestamp for PM database event
        """
        if circuit_type == 'RB':
            full_path = os.path.join(selected_path, 'pc/RB_Reference_PNO_b2.csv')
        elif circuit_type == 'RQ':
            full_path = os.path.join(selected_path, 'pc/RQ_Reference_PNO_b3.csv')
        else:
            raise ValueError('Circuit type {} not supported!'.format(circuit_type))

        ref_fpa_df = pd.read_csv(full_path)
        return ref_fpa_df[(ref_fpa_df['circuit'] == circuit_name)][col].values[0]

    # LEADS
    @staticmethod
    def read_current_leads_thresholds() -> pd.DataFrame:
        """Method reading a pandas DataFrame with Current Leads thresholds for RB circuit:

        :return: pd.DataFrame with reference Current Leads thresholds for RB circuit
        """
        full_path = os.path.join(selected_path, 'leads/RB_Reference_Threshold.csv')
        return pd.read_csv(full_path, index_col=0)

    # Feature Comparison
    @staticmethod
    def compare_features_to_reference(features_df: pd.DataFrame, circuit_type: str, system: str,
                                      *, wildcard=None, timestamp=None) -> pd.DataFrame:
        """ Method comparing input features with reference features given as a [min, max] range of values for:

        - Quench Heaters
        - Power Converter
        - Energy Extraction

        Method considers only a subset of reference features matching the input ones.

        :param features_df: input features to be compared to the references
        :type features_df: pd.DataFrame
        :param circuit_type:
        :type circuit_type: str
        :param system: system name
        :type system: str
        :param wildcard: wildcard in order to replace reference feature name
        :param timestamp: timestamp in order to include appropriate reference features (EE settings for RB were changed
        during the operation)

        :return: True if features are within min/max range from reference features, otherwise False
        """
        if system == 'QH':
            reference_df = Reference.read_quench_heaters_reference_features(circuit_type).T
        elif system == 'PC':
            reference_df = Reference.read_power_converter_reference_features(circuit_type).T
        elif system == 'EE':
            reference_df = Reference.read_energy_extraction_reference_features(circuit_type, timestamp).T
        else:
            raise KeyError('System {} is not supported'.format(system))

        # update wildcards in reference features
        if wildcard is not None:
            key, value = list(wildcard.items())[0]
            reference_df.columns = [col.replace('%{}%'.format(key), wildcard[key]) for col in reference_df.columns]

        features_df.index = ['act']
        common_columns = list(set(reference_df.columns.values).intersection(set(features_df.columns.values)))
        comparison_df = pd.concat([reference_df[common_columns], features_df[common_columns]], sort=False)
        comparison_df_t = comparison_df.T

        comparison_df_t['result'] = comparison_df_t.apply(
            lambda row: (row['act'] >= row['min']) and (row['act'] <= row['max']), axis=1)

        return comparison_df_t

    @staticmethod
    def compare_difference_of_features_to_reference(features_df: pd.DataFrame, reference_df: pd.DataFrame,
                                                    circuit_type: str, system: str, wildcard=None) -> pd.DataFrame:
        """ Method comparing the difference of the input features and the reference ones to an acceptance range

        :param features_df: input features
        :type features_df: pd.DataFrame
        :param reference_df: reference features
        :type reference_df: pd.DataFrame
        :param circuit_type: circuit type
        :type circuit_type: str
        :param system: system name
        :param wildcard: wildcard in order to replace reference feature name

        :return: a DataFrame with comparison of features
        """
        if system == 'QH':
            differences_df = Reference.read_quench_heaters_reference_differences(circuit_type).T
        else:
            raise KeyError('System {} is not supported'.format(system))

        if wildcard is not None:
            key, value = list(wildcard.items())[0]
            differences_df.columns = [col.replace('%{}%'.format(key), value) for col in differences_df.columns]

        features_df.index = ['act']
        reference_df.index = ['ref']

        comparison_df = pd.concat([reference_df[features_df.columns], features_df, differences_df[features_df.columns]],
                                  sort=False)
        comparison_df = comparison_df.T
        comparison_df['result'] = comparison_df.apply(lambda row: abs(row['act']-row['ref']) <= row['diff'], axis=1)

        return comparison_df

    @staticmethod
    def check_feature_comparison(comparison_df):
        """Method checking if the comparison dataframe contains all results within a prescribed range.
        In case the results fall outside of the prescribed range (some rows in the results column are False), then
        these rows are displayed.

        :param comparison_df: dataframe containing a result of a comparison in a result column
        :type comparison_df: pd.DataFrame

        :return: a DataFrame with comparison of features
        """
        if comparison_df['result'].all():
            return True
        else:
            print('The following parameters did not satisfy the comparison criteria:')
            if get_ipython() is None:
                print(comparison_df[comparison_df['result'] == False])
            else:
                display(HTML(comparison_df[comparison_df['result'] == False].to_html()))

            return False

==> lhcsmapi-1.4.58.tar.gz: lhcsmapi-1.4.58/lhcsmapi/reference/Reference.py <==
import os
from typing import List

import pandas as pd
from dateutil import parser
from IPython.display import display

from lhcsmapi.Time import Time


selected_path = os.path.dirname(__file__)


class Reference(object):
    """Reference class provides methods to access tables with reference values of components:

    - EE
    - Current Leads
    - MP3
    - Power Converter
    - QH
    - Query Table

    """

    # Query MP3 Quench Excel database columns
    @staticmethod
    def read_quench_database_columns(circuit_type: str, table_type='MP3') -> List[str]:
        """ Method reading columns of an MP3 quench database Excel file

        :param circuit_type: circuit type
        :type circuit_type: str
        :param table_type: table type (either MP3 or LHCSM)
        :type table_type: str

        :return: pd.DataFrame with names of columns for an MP3 quench database Excel file
        """

        full_path = os.path.join(selected_path, 'mp3/%s_%s_Quench_Database_Columns.csv' % (circuit_type, table_type))
        return pd.read_csv(full_path, index_col=0).columns

    # QH
    @staticmethod
    def read_quench_heater_reference_discharges(circuit_type: str) -> pd.DataFrame:
        """ Method reading reference quench heater discharge timestamps for a given circuit type

        :param circuit_type: circuit type
        :type circuit_type: str

        :return: pd.DataFrame with names of columns for full FPA quench analysis
        """

        full_path = os.path.join(selected_path, 'qh/{}_Reference_Discharges.csv'.format(circuit_type))
        return pd.read_csv(full_path, index_col=0)

    @staticmethod
    def get_quench_heater_reference_discharge(circuit_type: str, cell_or_circuit: str) -> int:
        """Method returning a timestamp corresponding to a reference quench heater discharge

        :param circuit_type:
        :type circuit_type: str
        :param cell_or_circuit: name of a magnet cell
        :type cell_or_circuit: str

        :return: int in a nanosecond precision unix timestamp of a reference QH timestamp
        """

        if circuit_type in ['RB', 'RQ']:
            col_name = 'Cell'
        elif circuit_type in ['IT', 'IPD2', 'IPD2_B1B2', 'IPQ2', 'IPQ4', 'IPQ8']:
            col_name = 'Circuit'
        else:
            raise ValueError('Circuit type %s not supported!' % circuit_type)

        qh_ref_discharges = Reference.read_quench_heater_reference_discharges(circuit_type)
        return qh_ref_discharges[qh_ref_discharges[col_name] == cell_or_circuit]["Timestamp"].values[0]

    @staticmethod
    def read_quench_heaters_reference_features(circuit_type: str) -> pd.DataFrame:
        """ Method reading a pandas DataFrame with reference values of QH features
        For all circuits:

        - initial voltage
        - final voltage

        :param circuit_type: circuit type
        :type circuit_type: str

        :return: pd.DataFrame with reference QH features
        """
        full_path = os.path.join(selected_path, 'qh/{}_Reference_Features.csv'.format(circuit_type))
        return pd.read_csv(full_path, index_col=0)

    @staticmethod
    def read_quench_heaters_reference_differences(circuit_type: str) -> pd.DataFrame:
        """ Method reading a pandas DataFrame with allowed differences of QH features from features calculated with
        the reference discharges.
        For all circuits (with QH voltage measurement):

        - voltage time constant

        For circuits with QH current measurement:

        - current time constant
        - initial resistance value

        :param circuit_type: circuit type
        :type circuit_type: str

        :return: pd.DataFrame with allowed differences of QH features from features calculated with the reference
        discharges.
        """
        full_path = os.path.join(selected_path, 'qh/{}_Reference_Differences.csv'.format(circuit_type))
        return pd.read_csv(full_path, index_col=0)

    # EE
    @staticmethod
    def read_energy_extraction_reference_features(circuit_type: str, timestamp_fgc='2021-01-01 00:00:00.000+01:00') \
            -> pd.DataFrame:
        """ Method reading a pandas DataFrame with min/max reference EE features
        RB:
        - R_1
        - R_2
        - t_delay_1
        - t_delay_2
        - tau_u_dump_res_1
        - tau_u_dump_res_2

        RQ:
        - R
        - t_delay
        - tau_u_dump_res

        :param circuit_type: circuit type
        :type circuit_type: str
        :param timestamp_fgc: FGC timestamp needed to select an appropriate EE reference features
        :type timestamp_fgc: str

        :return: pd.DataFrame with reference EE features
        """

        if circuit_type == 'RQ':
            full_path = os.path.join(selected_path, 'ee/RQ_Reference_Features.csv')
        elif circuit_type == 'RB':
            date_change = parser.parse('2018-01-01 00:00:00.000+01:00')
            timestamp_fgc_datetime = Time.to_datetime(timestamp_fgc)

            if timestamp_fgc_datetime > date_change:
                full_path = os.path.join(selected_path, 'ee/RB_Reference_Features_2018_01_01.csv')
            else:
                full_path = os.path.join(selected_path, 'ee/RB_Reference_Features_2014_01_01_2018_01_01.csv')
        else:
            raise KeyError('Circuit type {} not supported!'.format(circuit_type))

        return pd.read_csv(full_path, index_col=0)

    # PC
    @staticmethod
    def read_power_converter_reference_features(circuit_type: str) -> pd.DataFrame:
        """Method reading a pandas DataFrame with min/max reference features of a Power Converter:
        - tau_i_meas

        :param circuit_type: circuit type
        :type circuit_type: str

        :return: pd.DataFrame with reference Power Converter features
        """
        if circuit_type not in ['RB', 'RQ']:
            raise ValueError('Reference features for {} circuit type are not defined.'.format(circuit_type))

        full_path = os.path.join(selected_path, 'pc/{}_Reference_Features.csv'.format(circuit_type))
        return pd.read_csv(full_path, index_col=0)

    @staticmethod
    def get_power_converter_reference_fpa(circuit_type: str, circuit_name: str, col: str) -> str:
        """ Methor returning a timestamp of a reference FGC/EE timestamp for PM database event

        :param circuit_type: circuit type
        :type circuit_type: str
        :param circuit_name: circuit name
        :type circuit_name: str
        :param col: column: RB: fgcPm, eeEvenPm, eeOddPm, RQ: fgcPm, eePm
        :type col: str

        :return: str with a timestamp of a reference FGC/EE timestamp for PM database event
        """
        if circuit_type == 'RB':
            full_path = os.path.join(selected_path, 'pc/RB_Reference_PNO_b2.csv')
        elif circuit_type == 'RQ':
            full_path = os.path.join(selected_path, 'pc/RQ_Reference_PNO_b3.csv')
        else:
            raise ValueError('Circuit type {} not supported!'.format(circuit_type))

        ref_fpa_df = pd.read_csv(full_path)
        return ref_fpa_df[(ref_fpa_df['circuit'] == circuit_name)][col].values[0]

    # Feature Comparison
    @staticmethod
    def compare_features_to_reference(features_df: pd.DataFrame, circuit_type: str, system: str,
                                      *, wildcard=None, timestamp=None) -> pd.DataFrame:
        """ Method comparing input features with reference features given as a [min, max] range of values for:

        - Quench Heaters
        - Power Converter
        - Energy Extraction

        Method considers only a subset of reference features matching the input ones.

        :param features_df: input features to be compared to the references
        :type features_df: pd.DataFrame
        :param circuit_type:
        :type circuit_type: str
        :param system: system name
        :type system: str
        :param wildcard: wildcard in order to replace reference feature name
        :param timestamp: timestamp in order to include appropriate reference features (EE settings for RB were changed
        during the operation)

        :return: True if features are within min/max range from reference features, otherwise False
        """
        if system == 'QH':
            reference_df = Reference.read_quench_heaters_reference_features(circuit_type).T
        elif system == 'PC':
            reference_df = Reference.read_power_converter_reference_features(circuit_type).T
        elif system == 'EE':
            reference_df = Reference.read_energy_extraction_reference_features(circuit_type, timestamp).T
        else:
            raise KeyError('System {} is not supported!'.format(system))

        # update wildcards in reference features
        if wildcard is not None:
            key, value = list(wildcard.items())[0]
            reference_df.columns = [col.replace('%{}%'.format(key), wildcard[key]) for col in reference_df.columns]

        features_df.index = ['act']
        common_columns = list(set(reference_df.columns.values).intersection(set(features_df.columns.values)))
        comparison_df = pd.concat([reference_df[common_columns], features_df[common_columns]], sort=True)
        comparison_df_t = comparison_df.T

        comparison_df_t['result'] = comparison_df_t.apply(
            lambda row: (row['act'] >= row['min']) and (row['act'] <= row['max']), axis=1)

        return comparison_df_t

    @staticmethod
    def compare_difference_of_features_to_reference(features_df: pd.DataFrame, reference_df: pd.DataFrame,
                                                    circuit_type: str, system: str, wildcard=None) -> pd.DataFrame:
        """ Method comparing the difference of the input features and the reference ones to an acceptance range

        :param features_df: input features
        :type features_df: pd.DataFrame
        :param reference_df: reference features
        :type reference_df: pd.DataFrame
        :param circuit_type: circuit type
        :type circuit_type: str
        :param system: system name
        :param wildcard: wildcard in order to replace reference feature name

        :return: a DataFrame with comparison of features
        """
        if system == 'QH':
            differences_df = Reference.read_quench_heaters_reference_differences(circuit_type).T
        else:
            raise KeyError('System {} is not supported!'.format(system))

        if wildcard is not None:
            key, value = list(wildcard.items())[0]
            differences_df.columns = [col.replace('%{}%'.format(key), value) for col in differences_df.columns]

        features_df.index = ['act']
        reference_df.index = ['ref']

        comparison_df = pd.concat([reference_df[features_df.columns], features_df, differences_df[features_df.columns]],
                                  sort=False)
        comparison_df = comparison_df.T
        comparison_df['result'] = comparison_df.apply(lambda row: abs(row['act']-row['ref']) <= row['diff'], axis=1)

        return comparison_df

    @staticmethod
    def check_feature_comparison(comparison_df):
        """Method checking if the comparison dataframe contains all results within a prescribed range.
        In case the results fall outside of the prescribed range (some rows in the results column are False), then
        these rows are colored in red.

        :param comparison_df: dataframe containing a result of a comparison in a result column
        :type comparison_df: pd.DataFrame

        :return: a DataFrame with comparison of features
        """
        color = comparison_df['result'].map({False: 'background-color: red', True: ''})
        display(comparison_df.style.apply(lambda s: color))
        return comparison_df['result'].all()

    @staticmethod
    def get_mp3_period(date_time) -> str:
        """ Method returning an MP3 period name based on date and time

        :param date_time: date time for which MP3 period name is requested
        :return: method returns a string with MP3 period name
        """
        date_time_datetime = Time.to_string_short(date_time)
        full_path = os.path.join(selected_path, 'mp3/HWC_Operation_YETS_Periods.csv')
        periods_df = pd.read_csv(full_path)
        return periods_df[date_time_datetime <= periods_df['date_end']]['Period'].values[0]
